(* 
   File   : /home/cvs/cvsroot/projects/visualstate/scope/AbsynPrint.sml,v
   Comment: PrettyPrinting Absyn elements for testing and debugging
   Release: scope-0-10

     Copyright (C) 2001-2002 Andrzej Wasowski
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.



   AbsynPrint.sml,v 1.3 2002/10/24 20:17:09 wasowski Exp
*)

  open Haskellish Absyn Say Err
  infixr 0 $
  infix 9 sub

  val all = List.all

  (* some combinators *)

  fun lift_pr  (f :'a -> string) :'a -> unit = fn a => print (f a)
  fun lift_ln  (f :'a -> 'b    ) :'a -> 'b = fn a => f a before print "\n"

  (* and The real thing *)

  fun sprEBind (sy :symbol, pa :symbol list) : string = 
	String.concat (Symbol.name sy :: "(" 
		:: Typeset.c_params (map Symbol.name pa) [])

  fun sprTrigger (Event eb) : string = sprEBind eb
    | sprTrigger (EventGroup g) = Symbol.name g

  fun sprState (s :state) = Symbol.name (name s)
  val (sprStateor,sprStateand) = lower_orand sprState

  fun sprId (s :stateid) :string = Symbol.name s
  fun sprIds (ids :stateid list) = 
	String.concat $ Typeset.join_blanks(map sprId ids) []
  val prIds = lift_pr sprIds

  val sprInt = Int.toString
  fun sprInts (is :int list) = 
	String.concat $ Typeset.join_blanks(map sprInt is) []
  val prInts = lift_pr sprInts

  (* guard and action supressed *)

  fun sprPos (ss :andstate list) : string =
	String.concat("POS(" :: Typeset.join_blanks(map sprStateand ss) [")"])

  fun sprNeg (ss :andstate list) : string =
	String.concat( "NEG(" :: Typeset.join_blanks(map sprStateand ss) [")"])

  val prPos = lift_pr sprPos
  val prNeg = lift_pr sprNeg
  val prPosln = lift_ln prPos
  val prNegln = lift_ln prNeg
	
  fun sprTran ( {trigger,scope,pos,neg,guard,action,targets} :tran ) :string =
	String.concat (
	"[" :: sprStateor scope :: "] " :: sprTrigger trigger :: " : " 
		:: sprPos pos :: " " :: sprNeg neg :: " -> " 
			:: Typeset.join_blanks(map sprStateand targets) [])
			
  val prTran = lift_pr sprTran
  val prTranln = lift_ln prTran


