(* 
   File   : /home/cvs/cvsroot/projects/visualstate/scope/CodGenC1_LR.sml,v
   Comment: intermediate representation (mostly to be used by CodGenC1.
   Release: scope-0-10 
  
     Copyright (C) 2001-2002 Andrzej Wasowski
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   NOTES:

   All types used in intermediate representation are defined here.
   Data types only. No logics defined here (except from trivialities).

   CodGenC1_LR.sml,v 1.20 2002/11/15 22:45:08 wasowski Exp
*)

structure CodGenC1_LR = struct

(* some interface *)
open Types Exp CodGen

infix  3 --

structure SE = SymbolEnv

(* fundemental types imported *)

type symbol = Symbol.symbol
type ''a ulist = ''a UList.ulist
type 'a env = 'a SE.env
val  lookup = SE.lookup


(* Other types *)


(* events *)

type eenv = int env

(* runtime data layout markers (aka opcodes) *)

datatype marker = STMRK | HMRK | IMRK | FPSCOP | NFPSCOP 

(* number of markers (model dependent). We keep it constant but easily
   updateable by hand. To make it really dynamic and model dependent it
   should be moved to one of model representation/information records.
   (FIXME: this would be a nice thing to do in production version) *)

val markno :int = 5

(* low level representation byte code *)

datatype lratom = 
	  LRMrk          of marker	(* bytecode marker                  *)
	| LRAndStateRef  of symbol	(* reference to andstate by name    *)
        | LROrStateRef   of symbol	(* reference to orstate  by name    *)
	| LRNoEvent			(* a special "no event" event       *)
	| LREventRef     of symbol	(* reference to trans group by event*)
	| LRHistoryRef   of int		(* reference to a cell in history v.*)
	| LREnEx         of int * int	(* reference to entry/exit actions  *) 
	| LRAcGd         of int * int	(* reference to guard and action    *)
	| LRCondCounters of int * int	(* counter of pos and neg conds     *)
	| LRStateLabel   of symbol	(* mark start of state,no code emit *)
	| LREventLabel   of symbol	(* start of group of trans, no code *)

(* transitions *)	

type  stran = StaticScope.stran		(* intermediate form of transitions *)
type   tenv = lratom list env		(* group transitions by trigger     *)

(* variables *)

type vartype = Types.vartype
type vartypes= vartype env
type venv    = value env


(* events and signals *)

type signals = (lratom list) ulist


type evtypes = evtype env

val  sigsize = length

(* actions and guards *)

val  guardsize = fn _ => 1
type guard  = bexp list		(* unchanged from Absyn but deliberately      *)
				(* redefined                                  *)
type guards = guard ulist	(* where guards is a raw source guard list    *)
				(* from Absyn.sml                             *)
type guardref = int		
type guardrefs = guardref list

val always     :guard    = [Cst (CstI 1)]
val always_ref :guardref = markno


val  actionsize = fn _ => 1


type sexp = symbol * exp list		(* deliberately copied from Absyn.sml *)
					(* name, [param]                      *)

type action = { assgns  : assgn list,	(* copied from Absyn.sml, but         *)
		outputs : oexp list,    (* deliberately redefined             *)
		signals : sexp list }

type actions = action ulist	(* where actions is a raw source action 
				   from Absyn.sml *)

val null_action :action = { assgns = [], outputs = [], signals = [] }
val null_actref :int    = markno

(* information about exit types present *)

type exitmode = { pure : bool, impure : bool }

fun use_pure   mode = if #pure mode then mode
		      else { pure=true, impure= #impure mode }

fun use_impure mode = if #impure mode then mode
		      else { pure= #pure mode, impure = true }



(* system in intermediate representation *)

type lrsystem = { 

	name	 :string,		(* used for generating file name     *)
	
	hor	 :lratom list,		(* or part of hierarchy		     *)
	hand	 :lratom list,		(* and part of hierarchy	     *)
	root     :symbol,		(* name of topstate                  *)
	tenv     :tenv,			(* transitions array		     *)
	
	history  :lratom list,		(* list of local history vectors     *)
	signals  :signals,		(* list of triggering parts of acts  *)
	
	actions  :actions,		(* unique action bodies (ordered)    *)
	guards	 :guards,		(* unique guard bodies (ordered)     *)

	evtypes  :evtypes,		(* Unmodified from Absyn.sml         *)
	vartypes :vartypes,		(* Unmodified from Absyn.sml	     *)
	venv     :venv,			(* Unmodified from Absyn.sml	     *)
	cenv     :const env,		(* Unmodified from Absyn.sml         *)

	queuelen :int option,		(* only SOME if queue is used        *)
	statew	 :int,			(* upper bound of state configuration*) 
					(* size				     *)
	targetw  :int,			(* upper bound for number of targets *)
					(* in any single entering	     *)
	orsts    :int,			(* number of or-states		     *)
	andsts   :int,			(* number of and-states              *)
	exitmode :exitmode		(* exit mode used		     *)
}	 
 
(* type information about the model *)

type lrsizes = {	
	
	actionref :int,			(* inferred type of action refs      *)
	guardref  :int,			(* inferred type of guard ref type   *)
	orstref   :int,			(* inferred type of orstate refs     *)
	andstref  :int,			(* inferred type of andstate refs    *)
	signalref :int,			(* inferred return type for all acts *)
	evnumber  :int			(* events amount: id of last event+1 *)

}
(* scheduling related types *)

(* state addresses with automatic compression *)

datatype stdict = Direct   of int env
		| Indirect of int env * int list * int
		(* names to entries in dict, dict, no cells in dict *)

fun lookup_stdict (Direct env) = lookup env
  | lookup_stdict (Indirect (env,arr,size)) = lookup env

(* for debugging purposed *)

fun toList_stdict (Direct env) = SE.toList env
  | toList_stdict (Indirect (env,arr,size)) = SE.toList env

(* map events to groups of transitions *)

type tranidx = int env

(* type of data produced by scheduler *)

type scheduling = {

	eenv	  :eenv,		(* map event names to indices	     *)
	hor_dict  :stdict,		(* map or-states to addresses in hor *)
	hand_dict :stdict,		(* map andstates to addresses in hand*)
	tranidx   :tranidx              (* map events to position in transarr*)
	
}

type lrmodel = (lrsystem * lrsizes) list


(* some generic operations about packing and distributing integers over
   various integer sizes. USed in details of scheduling and compilation *)

(* split an integer number data into younger part of bytes length
   and older part - what is remaining. I think I should have put more effort
   in following packing functions. What I have now is big, slow and ugly. *)

local fun f 0 res = res -1
	| f n res = f (n-1) (256 * res)
in fun ones n = f n 1 end

fun partition_bits (data :int) (bytes :int) :int * int =
	let val mask   = Word.fromInt (ones bytes)
	    val wdata  = Word.fromInt data
	    val suffix = Word.andb(wdata,mask)
	    val prefix = Word.>>(wdata,Word.fromInt (bytes * 8))
	in (Word.toInt prefix, Word.toInt suffix) end;

fun dist chunk data 0 res = res
  | dist chunk data 1 res = data::res
  | dist chunk data n res = let val (pre,suf) = partition_bits chunk data 
			    in dist chunk pre (n-1) (suf::res) end

fun intgr (var :string) (cells :int) :codgen = 
	let fun f 0 = eoc
	      | f 1 = sg "+ (*((ptr)++))"
	      | f n = sg "+ (*((ptr)++)) << (sizeof(*(ptr)) * " 
				-- ig (n-1) -- sg ")" -- f (n-1)
	in cg [ "{ (", var, ") = " ] -- f cells -- sg "; }" end
(*
fun integr src_size tgt_size macro :codgen =
	if src_size >= tgt_size then fn os => "(d)"::os
	else let val mask = ones src_size
	     in 
		integr src_size (tgt_size - src_size) macro 
			o (fn os => " & d) + " ::itoa mask ::"(("::os) 
	     end
*)	

(* generic packing of integers to arrays of arbitrary size target chunks. *)

fun pack1 (data :int) (src_size :int) (tgt_size :int) :int list =
	let val chunks = Real.ceil((Real.fromInt src_size / Real.fromInt tgt_size))
	in dist tgt_size data chunks [] end;

fun pack2 (data1, data2) (src1,src2) (tgt_size) :int list = 
	if src1+src2 <= tgt_size then [(ones src1 + 1) * data2 + data1]
	else let val c1 = Real.ceil((Real.fromInt src1/Real.fromInt tgt_size))
		 val c2 = Real.ceil((Real.fromInt src2/Real.fromInt tgt_size))
		 val d2 = dist tgt_size data2 c2 []
		 val d1 = dist tgt_size data1 c1 d2
	in d1 end;

(* generic size estimation of packed data. Return number of target chunks *)

fun measure1 (src_size :int) (tgt_size :int) :int =
	Real.ceil((Real.fromInt src_size / Real.fromInt tgt_size)) 

fun measure2 (src1,src2) (tgt_size :int) :int =
	if src1+src2 <= tgt_size then 1
	else Real.ceil((Real.fromInt src1/Real.fromInt tgt_size)) +
	     Real.ceil((Real.fromInt src2/Real.fromInt tgt_size))

(* generation of generic macro accessing packed data. *)

(* only creates the body of macro. parameter is supposed to be called d *)

(*
fun access1 (src_size :int) (tgt_size :int) :codgen =
	let val 
*)


fun access2 (src1 :int, src2 :int) (tgt_size :int) :codgen =
	if src1+src2 <= tgt_size then
		cg ["{ (i) = *(ptr) & ", (Int.toString (ones src1)), "; ",
		    "j = *((ptr)++) >>  ", (Int.toString (src1 * 8)), "; }"]
	else let val c1 = Real.ceil((Real.fromInt src1/Real.fromInt tgt_size))
		 val c2 = Real.ceil((Real.fromInt src2/Real.fromInt tgt_size))
		 val m1 = intgr "i" c1 
		 val m2 = intgr "j" c2
	in sg "{ " -- m1 -- m2 -- sg " }" end

end
