(* 
   File   : /home/cvs/cvsroot/projects/visualstate/scope/CodGenC1_dump.sml,v
   Comment: Code dumper for CodGenC1 code generator
   Release: scope-0-10 

     Copyright (C) 2001-2002 Andrzej Wasowski
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   CodGenC1_dump.sml,v 1.32 2002/11/09 21:57:43 wasowski Exp
*)

structure CodGenC1_dump = struct

(* some interface *)

open Cppish Err CodGenC
infix 0 << <<< <<> <- (* some fun with C++ style IO *)
infix 3 --

type csystem = CodGenC1_back.csystem

(* file plate *)

val warn_generated = "/* This is a generated file! */\n/* Don't edit it manually.   */\n"

(* helper functions *)

fun prop (comment :string) (g :codgen) :codgen =
	cg ["/* ", comment, " */\n\n"] -- g -- sg "\n\n"

fun cond (absent :string) (present :string) (gopt :codgen option) :codgen = 
	case gopt of NONE   => cg ["/* ", absent, " */\n\n"]
		   | SOME g => prop present g

(* dump main c file *)

fun dump_system_c fbase (sys :csystem) :unit =
	let val os = TextIO.openOut (fbase ^ ".c")
	in 
		os << warn_generated << endl
		   
		   <- prop 
			"include public information" 
			(INCLUDE (fbase ^ ".h"))

		   <- prop
			"prototypes for functions used in actions/guards"
			(#externhdr sys)

		   <- prop
			"Mnemonics for LR-bytecode (used as comments) in arrays"
			(#lrmnemonics sys)

		   <- prop 
			"inferred type of action refs" 
			(#actionref sys)
		   <- prop
			"inferred type of guard refs"	
			(#guardref sys)
		   <- prop
			"inferred type of or-state ref"	
			(#orstref sys)
		   <- prop
			"inferred type of and-state ref"
			(#andstref sys)
		   <- prop
			"inferred return type for all actions"
			(#signalref sys)
		   <- prop
			"inferred type for refering to trans"
			(#tranref sys)
		   <- prop
			"type of iterators over trans (|traniter| >= |tranref|)"
			(#traniter sys)
		   <- prop 
			"type of configurations"
			(#conf sys)
		   <- prop
			"type for integer iterators over configuration arrays"
			(#confiter sys)
		   <- prop
			"type for integer iterators over sets of targets"
			(#tgtsiter sys)
		   <- cond
			"no iterators over queues (no queues) in this model"
			"type of integer iterators over signal queue elements"
			(#queueiter sys)
		   <- prop 
			"model-independent types (here for simplicity)"
			(#types sys)

		   <- cond 
			"type of symbols suppressed in release mode"
			"type of entries in symbolic dictionaries (debug)"
			(#symbol sys)

		   <- prop 
			"preprocessor constants for bytecode markers"
			(#markers sys)

		   <- prop
			"signal queue controlling macros and declarations" 
			(#queue sys)

		   <- prop 
			"variable definitions"
			(#vardefs sys) 

		   <- prop 
			"current event holder"
			(sg "struct StructEvent CurrEvent = { EVENTNO };\n\n")

		   <- prop 
			"history related control (switch and vector if needed)" 
			(#history sys)
			
		   <- prop 
			"prievious configuration" 
			(#prev_conf sys) 

		   <- prop 
			"next configuration"
			(#next_conf sys)

		   <- prop 
			"targets pool"
			 (#targets sys)

		   <- prop 
			"or-state projection of hierarchy"
			(#hor sys)

		   <- cond
			  "hor is accessed directly."
			  "dictionary for accessing hor" (#hor_dict sys) 

		   <- prop
			  "macro for accessing hor"
			  (#GET_OR sys)

		   <- prop 
			"and-state projection of hierarchy"
			(#hand sys)

		   <- cond 
			"hand is accessed directly."
			"dictionary for accessing hand" (#hand_dict sys)

		   <- prop
			"macro for accessing hand" 
			(#GET_AND sys)

		   <- prop
			"determine if model uses signal lists"
			(#SCOPE_SLIST sys)

		   <- prop
			"activate needed exit modes"
			(#exitmode sys)

		   <- cond 
			"no multiple signals on single transition"
			"groups of multiple signals on single transition" 
			(#signals sys)

		   <- prop
			"action dispatcher"
			(#actions sys)

		   <- cond
			"no predicative guards are used in this project"
			"guards dispatcher" (#guards sys)

		   <- prop 
			"transitions array"
			(#trans sys)

		   <- prop 
			"dictionary of transition lists"
			(#tranidx sys)

		   <- cond
			"symbolic names of or-states supressed in release mode"
			"symbolic names of or-states (debug mode)" 
			(#or_syms sys)

		   <- cond
			"symbolic names of and-states supressed in release mode"
			"symbolic names of and-states (debug mode)"
			(#and_syms sys)
		   <- cond 
			"symbolic names of events left out in release mode."
			"debug mode: symbolic event names" (#evdict sys)
			handle e => intErr e 200209267;

		os <- prop 
			"reference to topmost and-state"
			(#ROOT_STATE sys)

		   <- prop
			"bound on configuration size"
			(#STATE_WIDTH sys)

		   <- prop
			"bound on target width"
			(#TGT_WIDTH sys)

		   <- prop
			"macro for accessing encoded pos/neg counters"
			(#UNMANGLE_PCNC sys)

		   <- prop
			"macro for accessing encoded ac/gd references"
			(#UNMANGLE_ACGD sys)

		   <- prop
			"get enter action for state s"
			(#GET_ENTER sys)

		   <- prop
			"get exit action for state s"
			(#GET_EXIT sys)
		   <- prop
			"determine if arbitrary guards are used"
			(#SCOPE_GUARDS sys)

		   <- prop 
			"runtime interpreter"
			(INCLUDESTD "CodGenC1.c");


		TextIO.closeOut os
	end handle e => intErr e 200210060


(* dump external interface header *)

fun dump_system_h fbase (sys :csystem) :unit =
	let val os    = TextIO.openOut ( fbase ^ ".h")
	in 
		os << warn_generated << endl
		   <- prop 
			"fixed width integer types"
			(INCLUDESTD "VSTypes.h")
		
		(* model-independent but included here not to increase *)
		(* number of files involved.                           *)

		   << "/* API interface */\n\n"
		   << "void st_init(void);\n"
		   << "void macrostep(void);\n\n"

		   <- prop
			"inferred type for referring signals and events"
			(#eventtag sys)
	   
		   <- prop 
			"event names"
			(#evmacros sys) 
		handle e => intErr e 200209266;


		os <- prop 
			"event types"
			(#evtypes sys)
		   <- prop
			"union of event types"
			(#evunion sys)
		handle e => intErr e 200209268;
				     
		os <- prop
			"model constants"
			(#consts sys)
		
		   <- prop
			"variable declarations" 
			(#vardecls sys)
		handle e => intErr e 200210040;

		os   <- prop
			"number of events" 
			(#EVENTNO sys)
		handle e => intErr e 200210041;
	
		TextIO.closeOut os 
	end handle e => intErr e 200209269



fun dump_system (ofile :string option) (sys :csystem) :unit =
	let val fbase = case ofile of NONE   => #name sys 
				    | SOME f => f
	in 
		dump_system_h fbase sys;
		dump_system_c fbase sys
	end

end
