(* 
   File   : /home/cvs/cvsroot/projects/visualstate/scope/MkEnvStruct.sml,v 
   Comment: General efficient environments inspired by Andrew Appel's "Modern
            Compiler Implementation". Interface to environments inspired by
	    Peter Sestoft in PFOO course at IT-Universty, Spring 2002. 
   Release: scope-0-10

     Copyright (C) 2001-2002 Andrzej Wasowski
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   NOTES:

   The interface is similar to one proposed by Peter. Only key is no longer a
   type variable (fo equality type), but a fixed type (so it is dropped from
   many parameters). key may be any time for which toInt function may be
   given to the functor. 

   Structure HashEnv implements environments from Symbol.symbol to any type.
   See file HashEnv.sml

   FIXME: rename and split the files. There should be one called env-sig.sml.
   The other one should be called env-fn.sml and the functor should probably
   be called Environment (or Env).

   MkEnvStruct.sml,v 1.9 2002/10/24 20:17:10 wasowski Exp
*)   

signature ENV = sig

  type key
  type 'data env
  
  val empty    : 'data env
  val isEmpty  : 'data env -> bool
  val lookup   : 'data env -> key -> 'data 
  val count    : 'data env -> int
  val defined  : 'data env -> key -> bool
  val bind1    : 'data env -> key * 'data -> 'data env
  val bind     : 'data env -> (key * 'data) list -> 'data env
  val bindZip  : 'data env -> key list * 'data list -> 'data env
  val fromList : (key * 'data) list -> 'data env
  val toList   : 'data env -> (key * 'data) list  (* for test purposes *)
  val plus     : 'data env * 'data env -> 'data env
  val fullmap : (key * 'a -> 'b) -> 'a env -> 'b env
  val map     : ('a -> 'b) -> 'a env -> 'b env
end  


functor MkEnvStruct(s : sig type key
                            val  toInt: key -> int
			end ) :> ENV where type key = s.key =
struct

  structure I = Intmap
  type    key = s.key

  type 'a env = (key * 'a) I.intmap
  
  val empty = I.empty()

  (* yuck: horribly inefficient. I could make a wrapper for constant time, but
     I actually think this is a defficiency of Intmap which should provide it *)

  fun isEmpty env = I.numItems env <> 0
  
  val count = I.numItems

  fun lookup env k = 
	#2(I.retrieve(env, s.toInt k)) 
	handle NotFound => raise Subscript (* stupid: why have I done this? *)
		

  fun defined env k = case I.peek (env, s.toInt k) of
				  NONE   => false
				| SOME _ => true

  fun bind1 env (k,d) = 
	I.insert( env, s.toInt k, (k,d) )

  fun bind env lst = 
	foldl (fn ((k,d),e) => I.insert(e,s.toInt k,(k,d))) env lst

  fun bindZip env lsts = 
	ListPair.foldl (fn (k,d,e) => I.insert(e,s.toInt k,(k,d))) env lsts

  fun fromList lst = 
	bind (I.empty()) lst

  fun toList env = map #2 (I.listItems env)

  fun fullmap f  env = I.map (fn (i,(k,d)) => (k,f (k,d))) env

  fun map f env = fullmap (fn (k,d) => f d) env

  fun plus (env1,env2) =  
	let val lst2 = I.listItems env2
	in foldl (fn ((k,d),e) => I.insert(e,k,d)) env1 lst2 end
  
end


