(* 
   File   : /home/cvs/cvsroot/projects/visualstate/scope/Unparse.sml,v
   Comment: Save AST in vsr/vsp file back
   Release: scope-0-10

     Copyright (C) 2001-2002 Andrzej Wasowski
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   Unparse.sml,v 1.32 2002/10/24 20:17:10 wasowski Exp
*)

structure Unparse :> Unparse = struct

 open Typeset Say vsr 
 val name   = Symbol.name

 (* primitives *)

 fun pstpath   lst result : string list = join_dots (map name lst) result

 (* FIXME: pparams can be rephrased more efficiently with foldr with no concat *)
 fun pparams   lst result : string list = c_params (map (fn l => concat(pstpath l [])) lst) result
 fun pc_params lst result : string list = c_params (map name lst) result;


 fun addExt ext fpath = Path.joinBaseExt { base = Path.base fpath, 
					   ext = SOME ext }

 fun pinstance( id, SOME params ) result = name id :: " " :: pparams params result
   | pinstance( id, NONE        ) result = name id :: result
 
 (* FIXME: intending tabs is currently done in completely stupid way in
    pfield in ptag and in dropComma *)

 fun dropComma (        ","::lst) = lst
   | dropComma (",\n\t\t\t"::lst) = lst
   | dropComma (             lst) = lst


 fun ptag ( NONE     ) result : string list = result
   | ptag ( SOME lst ) result = 
	let 
	      fun mkv value result = " = \"" :: value :: "\"" :: result 
	      fun pfield (PlainTag (fname, value)) result = 
			",\n\t\t\t" :: fname :: mkv value result
		| pfield (SysName  (n,s)) result =
			",\n\t\t\t"::"VS_SYSTEM_" :: Int.toString n :: "_NAME" :: mkv (name s) result
			
		| pfield (FileName (n,m,s)) result =
			",\n\t\t\t"::"VS_SYSTEM_" :: Int.toString n :: "_FILE_" 
			:: Int.toString m :: "_NAME" :: mkv (name s) result
			
		| pfield (FilePath(n,m,s)) result =
			",\n\t\t\t"::"VS_SYSTEM_" :: Int.toString n :: "_FILE_"
			:: Int.toString m :: "_PATHNAME" :: mkv (addExt "vsr" (name s)) result
			
		| pfield (CallTag (n,m,i)) result =
			",\n\t\t\t"::"VS_SYSTEM_" :: Int.toString n :: "_FILE_" 
			 :: Int.toString m :: "_CALL" :: mkv (concat (pinstance i [])) result

		| pfield (Explanation  e) result = 
			",\n\t\t\t"::"Explanation" :: mkv (name e) result

		| pfield (ProjectExpl  s) result = 
			",\n\t\t\t"::"VS_PROJECT_EXPL":: mkv (name s) result

		| pfield (ProjectName  s) result = 
			",\n\t\t\t"::"VS_PROJECT_NAME":: mkv (name s) result

		| pfield (SystemExpl (n,s)) result = 
			",\n\t\t\t"::"VS_SYSTEM_" :: Int.toString n :: "_EXPL":: mkv (name s) result

		| pfield (FileExpl (n,m,s)) result = 
			",\n\t\t\t"::"VS_SYSTEM_" :: Int.toString n :: "_FILE_" 
			 :: Int.toString m :: "_EXPL" :: mkv (name s) result

		| pfield (FreeTextText (n,s)) result = 
			",\n\t\t\t"::"FreeTextText" :: Int.toString n :: mkv (name s) result

		| pfield (SystemNote (n,m,s)) result =
			",\n\t\t\t"::"System_" :: Int.toString n :: "_" :: Int.toString m
			:: "_NoteText" :: mkv (name s) result
			
		| pfield (CoderCmd (n,s)) result = result 
		(* FIXME CodeCmd should only be omitted in anonymizing mode *)

		| pfield (CenterHeader s) result = 
			",\n\t\t\t"::"CenterHeader" :: mkv (name s) result
		| pfield (RightHeader  s) result = 
			",\n\t\t\t"::"RightHeader"  :: mkv (name s) result
		| pfield (LeftFooter   s) result = 
			",\n\t\t\t"::"LeftFooter"   :: mkv (name s) result
		| pfield (CenterFooter s) result = 
			",\n\t\t\t"::"CenterFooter" :: mkv (name s) result
		| pfield (RightFooter  s) result = 
			",\n\t\t\t"::"RightFooter"  :: mkv (name s) result
		| pfield (ActionFile   s) result = 
			",\n\t\t\t"::"ActionFile"   :: mkv (name s) result

	in "\n\t\t\t[[ " :: dropComma( foldr (fn (t,r) => pfield t r) (" ]]"::result) lst) end
	
 (* printing directives *)
 
 fun pdir (Dir_visualSTATE s) result = s :: result
   | pdir (Dir_Date        s) result = s :: result
   | pdir (Dir_Time        s) result = s :: result 
   | pdir (Dir_Field       s) result = "@Field" :: ptag s result

 fun pdirs tabs result = 
	vsr_special_block_trans "@Directives" pdir tabs result;

 (* printing Types *)

 fun ptype (ty :simplety) (result :string list) :string list = 
	Types.tyname ty::result 

  (* printing literals *)

  fun pliteral (CharConst s) = "'" ^ s ^ "'"
    | pliteral (DecConst  i) = Int.toString i  
    | pliteral (HexConst  i) = "0x" ^ Int.fmt StringCvt.HEX i
    | pliteral (RealConst s) = s

 (* printing expressions *)

 val pbinop = CodGenC.dump_Prim2

 val punop  = CodGenC.dump_Prim1

 fun pvar  ( SimpleVar s          ) result = name s :: result 
   | pvar  ( SubscriptVar(s,e    )) result = name s :: "[" :: pexpr e ("]"::result)

 and pexpr ( ExprLiteral l        ) result = pliteral l :: result
   | pexpr ( ExprVar     v        ) result = pvar v result
   | pexpr ( ExprFunCall f        ) result = pfunCall f result
   | pexpr ( ExprPar     a        ) result = "(" :: pexpr a ( ")" :: result )
   | pexpr ( ExprUnary   (ope,  a)) result = punop ope :: pexpr a result
   | pexpr ( ExprBinary  (a,ope,b)) result = pexpr a (pbinop ope :: pexpr b result)

 and pfunCall (FunCall  (s,p    ) ) result = 
	name s :: " (" :: dropComma
		( foldr (fn (e,r) => "," :: pexpr e r) (")"::result) p )

 (* printing records & events & signals *)	
 
 fun ptyping (t,i) result = ptype t (" " :: name i :: result)

 fun prec ( id, NONE        ) result = name id :: result
   | prec ( id, SOME []     ) result = name id :: " ()" :: result
   | prec ( id, SOME (h::l) ) result = name id :: " (" :: ptyping h 
					(foldr (fn (t,r)=> ","::ptyping t r) (")"::result) l)

 fun pevent ( r, tag   ) result = prec r (ptag tag result)

 fun pevents  tabs evts result = 
	vsr_block_trans "@ExternalEvent" pevent tabs evts result
 fun psignals tabs evts result = 
	vsr_block_trans "@Signal"        pevent tabs evts result

 fun peventInstance (id, NONE   ) result = name id :: result
   | peventInstance (id, SOME ps) result = name id :: " " :: pc_params ps result

 (* printing functions and actions *)
 
 fun pfunc ( ty, id, params ) result = ptype ty  (" " :: prec (id, params) result)

 fun pactionSpec ( ActionSpec      (f,t)) result = pfunc f (ptag t result)
   | pactionSpec ( ActionTimerSpec (f,t)) result = "@Timer " :: pfunc f (ptag t result)

 fun pactionSpecs tabs acts result 
	= vsr_block_trans "@ExternalAction" pactionSpec tabs acts result

 (* printing int/ext variable declarations *)

 fun prange ( SOME(l,r) ) result = "{":: pliteral l :: "," :: pliteral r :: "} " :: result
   | prange ( NONE      ) result = result

 val parray = bracket o pliteral 

 fun pvarinit ( SOME l ) = " = " ^ pliteral l
   | pvarinit ( NONE   ) = ""

 fun parrinit ( SOME l ) result = " = {" :: join_commas (map pliteral l) ("}"::result)
   | parrinit ( NONE   ) result = result

 fun pvarDecl ( VarSimple (ty, id,    r, i, t) ) result = 
	ptype ty (" " :: name id :: " " :: prange r (pvarinit i::ptag t result))
   | pvarDecl ( VarArray  (ty, id, d, r, i, t) ) result =
        ptype ty (" " :: name id :: " " :: prange r (parray d::parrinit i (ptag t result)))

 fun pextvars tabs vars result = 
	vsr_block_trans "@ExternalVariable" pvarDecl tabs vars result
 fun pintvars tabs vars result = 
	vsr_block_trans "@InternalVariable" pvarDecl tabs vars result
 

 (* printing constants *)

 fun pconst ( id,ini,t ) result = name id :: pvarinit (SOME ini) :: ptag t result 

 fun pconsts tabs consts result = 
	vsr_block_trans "@Constant" pconst tabs consts result



 (* -- Towards internal interface ----------- *)

 (* printing event groups *)

 fun pmembers (SOME l) result = " (" :: join_commas (map name l) (")"::result)
   | pmembers (NONE  ) result = result

 fun pevtgrp (id,members,tag) result = name id :: pmembers members (ptag tag result)

 fun pevtgrps tabs lst result = 
	vsr_block_trans "@EventGroup" pevtgrp tabs lst result

 (* Printing state declarations *)
 
 fun psttype InitialState     = "@Initial "
   | psttype HistoryState     = "@History "
   | psttype DeepHistoryState = "@DeepHistory "
   | psttype FinalState       = "@Final "
   | psttype NormalState      = ""


 fun pstate { stateType, name=id, refersTo=SOME i, tag    } result = 
	psttype stateType::name id::" = "::pinstance i (ptag tag result)
   | pstate { stateType, name=id, refersTo=NONE, tag, ... } result = 
	psttype stateType :: name id :: ptag tag result 


 fun pstates tabs states result = 
	vsr_block_trans "@State" pstate tabs states result

 (* printing state regions *)

 fun pregion (id, NONE,      tag) result = name id :: ptag tag result
   | pregion (id, SOME pars, tag) result = name id :: " " :: pc_params pars (ptag tag result)

 fun pregions tabs regions result = 
	vsr_block_trans "@StateRegion" pregion tabs regions result

 (* printing actions *)

 fun paction ( ActionNextState s          ) result = pstpath s result
   | paction ( ActionFunCall   f          ) result = "#"  :: pfunCall f result
   | paction ( ActionSignal   (s, NONE)   ) result = "^"  :: name s :: result
   | paction ( ActionSignal   (s, SOME p) ) result = "^"  :: pfunCall (FunCall(s,p)) result
   | paction ( ActionAssign   (v,e,true ) ) result = 
	"[|" :: pvar v ("=" :: pexpr e ("|]"::result))
   | paction ( ActionAssign   (v,e,false) ) result = 
	"[(" :: pvar v ("=" :: pexpr e (")]"::result))

 fun pactions lst result = foldr (fn (a,r) => " " :: paction a r) result lst
 
 (* printing guards *)

 fun pguard (GuardPos s         ) result = pstpath s result
   | pguard (GuardNeg s         ) result = "!" :: pstpath s result
   | pguard (GuardExpr (e,true )) result = "[| " :: pexpr e (" |]"::result)
   | pguard (GuardExpr (e,false)) result = "[( " :: pexpr e (" )]"::result)

 fun pguards lst result = foldr (fn (g,r) => " " :: pguard g r) result lst
 



 (* printing rules *)

 fun perule (actions, tag) result = pactions actions (ptag tag result)

 fun prule {event, condition, actions, tag} result =
	peventInstance event (pguards condition (" : " :: pactions actions (ptag tag result)))

 fun pentry tabs rules result = 
	vsr_block_trans "@Entry" perule tabs rules result;
	
 fun pexit     tabs rules result = 
	vsr_block_trans "@Exit" perule tabs rules result;
	
 fun pintrules tabs rules result = 
	vsr_block_trans "@InternalRule" prule  tabs rules result;

 fun ptrsrules tabs rules result = 
	vsr_block_trans "@TransitionRule" prule  tabs rules result;

 fun pdorule (i, tag) result = pinstance i (ptag tag result)

 fun pdorules tabs rules result = 
	vsr_block_trans "@Do" pdorule tabs rules result




 (* printing state machines (including top state machine) *)

 fun pmparams []       result = result
   | pmparams (h::lst) result = 
		"(" :: " @State " :: name h ::  
		foldr (fn (p,r) => ", @State " :: name p :: r) (")"::result) lst
 

 fun pmachinedef tabs keyword { name=id, param,  
				interface = {events, actions, extvars,         
					     intvars, constants, groups, 
					     signals },
				regions, entryRules,exitRules, doRules,
				intRules, trsRules, states } result =
	let 
		val tabs' = indin tabs	
	in
		  tabs :: keyword :: " " :: name id :: pmparams param
		   (" = " :: "\n\n" :: pevents tabs' events
		    ("\n" :: pactionSpecs tabs' actions 
		     ("\n" :: pextvars tabs' extvars 
		      ("\n" :: pintvars tabs' intvars
		       ("\n" :: pconsts  tabs' constants
			("\n" :: pevtgrps tabs' groups
			 ("\n" :: psignals tabs' signals
			  ("\n" :: pstates  tabs' states
			   ("\n" :: pregions tabs' regions
			    ("\n" :: pentry   tabs' entryRules
			     ("\n" :: pexit    tabs' exitRules
			      ("\n" :: pdorules tabs' doRules
			       ("\n" :: pintrules tabs' intRules
				("\n" :: ptrsrules tabs' trsRules 
				 ("\n" :: tabs :: "@End" :: result
				  )))))))))))))))
	end
		

 fun pmachine tabs (StateMachine (m,tag)   ) result = 
	pmachinedef tabs "@StateMachine" m 
		(ptag tag (";\n\n"::result))
   | pmachine tabs (TopStateMachine (m,tag)) result = 
	pmachinedef tabs "@TopStateMachine" m 
		(ptag tag (";\n\n"::result))

 fun pmachines tabs mlst result = 
	foldr (fn (m,sl)=>pmachine tabs m sl) result mlst




 (* printing interface *)

 fun pinterface {events, actions, extvars, constants, 
                 intvars, groups, signals} result =
	"@ExternalInterface\n\n" :: pevents "\t" events 
	("\n" :: pactionSpecs "\t" actions 
	("\n" :: pextvars     "\t" extvars
	("\n" :: pconsts      "\t" constants 
	("\n@InternalInterface\n\n" :: pintvars "\t" intvars
	("\n" :: pevtgrps     "\t" groups    
	("\n" :: psignals     "\t" signals
	("\n" :: result )))))))
	




 (* printing the main program *)	

 fun pvsr { directives=dirs, interface, machines, 
                 top, filepath } = 

	"@visualSTATE @RuleBase\n\n" :: pdirs "" dirs 
		("\n" :: pinterface interface 
			("@Implementation\n\n" :: pmachines "\t" machines 
				("\n" :: pmachines "\t" [top] ["\n"] )))
			
 fun pvsp { directives=dirs, interface, sysnames, calls, 
		 files, programs, filepath } = 
	let 
		val _ = app (save o Program o #3) programs
	in
		"@visualSTATE @Project\n\n"
		:: pdirs "" dirs ("\n"::pinterface interface [])
	end

 and save (Project (p as { filepath,...})) =
	let 
		val fname = addExt "vsp" (Path.file (name filepath))
		val os    = TextIO.openOut fname
		val tks   = pvsp p
	in
		app (fn s => TextIO.output (os,s)) tks;
		TextIO.closeOut os;
		sayln VMed ( "> " ^ fname )
	end
   | save (Program (p as { filepath,...})) =
	let 
		val fname = addExt "vsr" (Path.file (name filepath))
		val os  = TextIO.openOut fname
		val tks = pvsr p
	in
		app (fn s => TextIO.output (os,s)) tks;
		TextIO.closeOut os;
		sayln VMed ("> " ^ fname)
	end

end


